
**************************************
* Title: Composer's Workbench
* Author: Matt Munz
* Date: 12/9/5
**************************************

**************************************
Changelog
**************************************
12/17/4 Submit as final project for FP class
9/30/5  Continuting life in independent research project into CA Etudes
10/1/5  Removing dependence on QuickCheck
11/15/5 Added extractPitches and transposeM
12/9/5  Moved generic functions to Miscellany.lhs
        Added extractDurations
12/17/5 Added sequencePhrases

**************************************
Todo
**************************************
Extract the library functions and example uses into separate modules

**************************************
Abstract
**************************************

> module ComposersWorkbench where
>
> import List hiding (transpose)
> import Ratio
> import Array
> import Haskore
> import System
> import Miscellany

  The Composer's Workbench is a collection of functions that may be useful for music analysis and composition.  This document is written in the Literate Programming style.  I will attempt to include a reasonable amount of documentation and background material.  Some Music Theory sources are quoted extensively.  A bibliography appears at the end of this text.

********
* Index
********

Introduction
Basic Types
Generic Pitch Class Functions
Generic Music Functions
Prime Form
Proof of a Property of Prime Forms
Interval Vectors
Rhythm
An Example
Mapping Text to Pitch
Generic Functions
Future Directions
References 

***************
* Introduction
***************

  The Problem:

    The ideal composition is a direct and clear expression of the composer's artistic vision.  When the audience can hear what the composer "hears" in her mind, her musical expression (or at least the first stage of it) is complete.  Of course, in practice, the dialogue between the composer and the audience is mediated in many ways.  Between the formulation of musical ideas and the final performance, there may be pen and paper, conductors, instrumentalists, singers, etc., all which impact the composer's expression in various ways.  As such, much of the composer's time is spent (and rightly so) in managing the medium of her expression.  Consider transposition, for example -- a tedious task that is necessary to assure correct performance, but often has little to do with the composer's artistic goals. After transposition, the music sounds no different to the audience.  It is only done to make the piece easier to perform.

  The Solution:

    Wherever possible, use information technology to create an environment in which the composer may focus on the task of generating novel musical ideas.  It might be illustrative to consider the extreme case, where there exists some device which directly connects the composer's brain to the listener's brain, and the composer need only think the sounds into existence for the listener to "hear" them.  Of course, such a device exists only in science fiction, but we can nonetheless start to approach a clearer musical dialogue through existing technology.  To further this aim, a DSL and, time permitting, a GUI, will be implemented in Haskell to serve as a component of the composer's work environment. 

    Given that Music is a less-formalized, less-known domain, as compared to, say, Mathematics, it will be necessary to actively explore the domain in search of algorithms and abstractions that may be suitable for translation to / automation in Haskell.  The following three approaches will be used to cull this information from the field.  
    
      1) Observe the composer at work
      2) Automate existing algorithms in the domain (the "rules" of composition in various styles)
      3) Build a "vocabulary" for analysis, and then "invert" it for compositional use

    While, in the long term, #1 is likely the most promising approach (it may in fact subsume the other two approaches), in the short term, #2 and #3 will be more productive, since there is a large body of musicological and music theoretic literature available that can be used to quickly populate the workbench with music transformation and analysis algorithms.  

    In the following, I have precisely followed this short-term strategy.  Some techniques of modern atonal analysis are transcribed, and then used to analyze a simple piece, Bartok's Mikrokosmos #37.  Finally, the products of this analysis are used to synthesize an origional composition.  Other "goodies" are included here, to stimulate the imagination, including a piece derived from text-pitch mappings.  

    Ultimately, the creation of a conceptually-consistent DSL for music composition, suitable for composition at a high level of abstraction, and usable in every-day practice for composers, is beyond the scope of this project.  In the "Further Directions" section I outline some of the ways in which Haskore might be modified/extended to cleanly integrate the techniques and ideas explored here. 

**************
* Basic Types
**************

  A few simple types are sufficient for performing basic atonal analysis.

  "Pitch Classes
    * Any two pitches which sound the same on an equal tempered scale (for example, C and B) or are only different due to octave displacement are said to belong to the same "pitch class".
    * Pitch classes can also be numbered:  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11.
      These numbers are sometimes called "Pitch Class Representatives".
  Pitch Class Sets
    * A "Pitch Class Set" is a list of pitch class numbers:   [0, 4, 7, 10]   (note the square brackets)
    * In Pitch Class sets, octave doublings and displacements are ignored:
          o [0, 3, 7, 12]  =>  [0, 3, 7]    
          o [14, 7, 11]  =>  [2, 7, 11]
  Transposing Pitch Class Sets
    * To transpose a pitch class set, add (or subtract) the same number to all elements of the list:
      [0,1,4]  =>  (transpose up a major third)  [0+4, 1+4, 4+4]  =>  [4,5,8]
  Inverting Pitch Class Sets
    * To invert a PC Set, subtract each element of the list from 12:
      [0,1,4]  =>  [12 - 0, 12 - 1, 12 - 4]  =>  [12, 11, 8]  =>  [0, 11, 8]
    * By convention, simple inversion is always around Pitch Class C (0). Therefore, any note of the chord which is N half-steps above C, will be flipped to be come a note N half-steps below C. In the above example, the note "E" (4 half-steps above C) was flipped to become "A" (4 half-steps below C)."

  - Paul Nelson 

> type PitchClassRep = Int
> type PitchClassSet = [PitchClassRep]
> type IntervalVector = Array Int Int -- length == 6.  Interval Vectors are described later.
> type PitchClassSetTransform = PitchClassSet -> PitchClassSet
> type MusicTransform = Music -> Music

**************************************
* Generic Pitch Class Set Functions
**************************************

The normalize function assures that the PitchClassRep stays "in bounds" (0-12)

> normalize :: PitchClassRep -> PitchClassRep
> normalize pcRep = pcRep `mod` 12

> normalizePCS :: PitchClassSetTransform
> normalizePCS = map normalize

> transpose :: Int -> PitchClassSetTransform
> transpose interval = normalizePCS . (map (+ interval))

> transposeB :: Int -> PitchClassSetTransform
> transposeB interval = map (+ interval)

> invertPCS :: PitchClassSetTransform
> invertPCS = normalizePCS . (map (12 -))

> transposeInvert :: Int -> PitchClassSetTransform
> transposeInvert interval = (transpose interval) . invertPCS

maxInterval finds the largest interval between any two pitch classes in the set.

> maxInterval :: PitchClassSet -> Int
> maxInterval pcs = (listMax pcs) - (listMin pcs) 

> transpose20H :: PitchClassSetTransform
> transpose20H []         = []
> transpose20H pcs@(pc:_) = transpose (negate pc) pcs

> multiplyPCS :: Int -> PitchClassSetTransform
> multiplyPCS factor = normalizePCS . (map (factor *))

> interpolate :: PitchClassSet -> PitchClassSet -> PitchClassSet
> interpolate []         pcs        = pcs
> interpolate pcs        []         = pcs
> interpolate (pc1:pcs1) (pc2:pcs2) = pc1:pc2:(interpolate pcs1 pcs2)

cycleIn reorders the pitches from the outside in.  A Boolean is used to choose the method of interpolation.

> cycleIn :: Bool -> PitchClassSetTransform 
> cycleIn flag pcs 
>   = case flag of
>       True  -> interpolate half1 half2 
>       False -> interpolate half2 half1 
>     where
>       halve aList = splitAt (floor ((fromInt (length aList)) / 2)) aList
>       halves      = halve pcs
>       half1       = fst halves 
>       half2       = reverse $ snd halves 

> extractIntervals :: PitchClassSet -> [Int]
> extractIntervals [] = []
> extractIntervals pcs@(_:pcs') = zipWith makeInterval pcs pcs' where
>   makeInterval pc1 pc2 = pc2 - pc1

> intervals2PCS :: PitchClassRep -> [Int] -> PitchClassSet
> intervals2PCS pc [] = [pc]
> intervals2PCS pc (interval:intervals) = (pc:(intervals2PCS (pc+interval) intervals))

contractIntervals modifies the interval "breadth" of a PCS, while preserving its "contour".

> contractIntervals :: Ratio Int -> PitchClassSetTransform
> contractIntervals scale pcs@(pc:_) = intervals2PCS pc intervals where
>   intervals :: [Int]
>   intervals = map (floor.toRational.(scale*).fromInt) (extractIntervals pcs)

The following functions convert from PCS to Music.

> pcs2Chord :: PitchClassSet -> Music
> pcs2Chord pcs = foldr (:=:) emptyMusic (map pc2Note pcs)

> pc2Note pc = Note (pitch (pc + (12*5))) 1 []

> pcs2Melody :: PitchClassSet -> Music
> pcs2Melody pcs = foldr (:+:) emptyMusic (map pc2Note pcs)

> makeMelody :: Int -> PitchClassSet -> Rhythm -> Music
> makeMelody interval (pc:pcs) (NoteR dur)
>   =  (Note (pitch (pc + interval)) dur []) 
> makeMelody interval _ (RestR dur) = (Rest dur) 
> makeMelody interval pcs (r1 :&: r2) 
>   = (makeMelody interval (take (numNotes r1) pcs) r1) :+: 
>     (makeMelody interval (drop (numNotes r1) pcs) r2)

***************************
* Generic Music Functions
***************************

  An "empty" constructor is useful for generic operations like fold.  This function approximates that feature.

> emptyMusic :: Music
> emptyMusic = Rest 0

Extract a linear set of pitches from the music, in order.  

> extractPitches :: Music -> [Pitch]
> extractPitches m = case m of
>    (Note p _ _)  -> [p]
>    (m0 :+: m1)   -> extractPitches m0 ++ extractPitches m1
>    (_ :=: _)     -> error "polyphony not yet supported"
>    (Trans _ _)   -> error "Trans constructor not yet supported"
>    (Rest _)      -> []
>    (Tempo  _ m0) -> extractPitches m0
>    (Instr  _ m0) -> extractPitches m0
>    (Player _ m0) -> extractPitches m0
>    (Phrase _ m0) -> extractPitches m0

Extract a sequence of note durations from a linear piece of music, in order.  

> extractDurations :: Music -> [Dur]
> extractDurations m = case m of
>    (Note _ d _)  -> [d]
>    (m0 :+: m1)   -> extractDurations m0 ++ extractDurations m1
>    (_ :=: _)     -> error "polyphony not yet supported"
>    (Trans _ m0)   -> extractDurations m0
>    (Rest _)      -> []
>    (Tempo  _ m0) -> extractDurations m0
>    (Instr  _ m0) -> extractDurations m0
>    (Player _ m0) -> extractDurations m0
>    (Phrase _ m0) -> extractDurations m0

Transpose all pitches in music by amount.  (Does not use the Trans constructor, also "compiles away" trans constructors along the way)

> transposeM :: Int -> Music -> Music
> transposeM i m = case m of
>    (Note p d as)  -> Note (trans i p) d as
>    (m0 :+: m1)    -> transposeM i m0 :+: transposeM i m1
>    (m0 :=: m1)    -> transposeM i m0 :=: transposeM i m1
>    (Trans i' m0)  -> transposeM (i+i') m0
>    (Rest d)       -> Rest d
>    (Tempo  ri m0) -> Tempo ri (transposeM i m0)
>    (Instr  x  m0) -> Instr x (transposeM i m0)
>    (Player pn m0) -> Player pn (transposeM i m0)
>    (Phrase as m0) -> Phrase as (transposeM i m0)

Sequence a set of phrases into one piece of music

> sequencePhrases :: [Music] -> Music
> sequencePhrases m = foldr (:+:) emptyMusic (map (Phrase []) m)

*************
* Prime Form
*************

  Prime Form calculation is a staple of atonal analysis.

  "Similar Pitch Class Sets:  Set Classes & Prime Forms
    * If two PC Sets differ only by transposition or inversion, then they belong to the same Set Class.
    * There are only 208 different Set Classes.
    * Each Set Class is represented by a "Prime Form" PC Set. For example:
      [0,1,4]; [3,4,7]; [0,3,4]; [5,8,9]; and [8,9,0]  all belong to the Prime Form:  (0,1,4)
  For example, the prime form of [0,8,6,8] is (0,2,4,8); and the prime form of [2,4,8,9] is (0,1,5,7)"

  - Paul Nelson 

Here's my algorithm for finding the prime form of a given set.

1) put chord in ascending order
1.1) transpose chord so that start is 0 (here on called transpose20H)
2) get all rotations of chord where a rotation is x,y,z -> y,z,x+12
   e.x.) [0,4,7] rotation set is [0,4,7],[4,7,12],[7,12,16]
3) transpose20H all rotations
  e.x.) [0,4,7],[0,3,8],[0,5,9]
4) the one whose last element is smallest is the winner -- e.x.) [0,4,7]
5) if there are two rotations w/ the same last element, start at the head of each and pairwise compare them.  the first one smaller is the loser

e.x.) applied to [6,8,0,4]
1) [0,4,6,8]
2) [0,4,6,8], [4,6,8,12],[6,8,12,16],[8,12,16,18]
3) [0,4,6,8],[0,2,4,8],[0,2,6,10],[0,4,8,10]
4) [0,4,6,8],[0,2,4,8]
5) [0,2,4,8]

  According to Gary R. Tucker, there are two (slightly) different methods for computing the prime form of a PC Set. (see References)  The method I have used involves "packing to the left", as it is apparently the origional method employed by Allen Forte.  The "packed to the right" method differs from this in exactly 5 cases.  Since the interval content of the set is identical to its left-packed counterpart, this likely doesn't amount to a big difference either way.

> primeForm :: PitchClassSetTransform 
> primeForm pcs = mostPacked (normalForm0 pcs) (normalForm0 (invertPCS pcs))

> normalForm0 pcs = resultStep5 where
>   resultStep1 = transpose20H (sort pcs)
>   resultStep4 = collectSmallest (map transpose20H (allRotations resultStep1))
>   resultStep5 = if ((length resultStep4) == 1)
>	          then head resultStep4
>	          else findMostPacked (head resultStep4) (tail resultStep4)

> collectSmallest :: [PitchClassSet] -> [PitchClassSet]
> collectSmallest (pcs':pcss) 
>   = collSmallestIter (last pcs') [pcs'] pcss where
>   collSmallestIter :: PitchClassRep -> [PitchClassSet] -> [PitchClassSet] -> [PitchClassSet]
>   collSmallestIter _   matches [] 
>     = matches
>   collSmallestIter minV matches (pcs':pcss) 
>     = collSmallestIter' minV matches pcs' pcss (last pcs')
>   collSmallestIter' minV matches pcs' pcss lastPCS'
>     | (lastPCS' < minV)  = collSmallestIter lastPCS' [pcs']         pcss
>     | (lastPCS' == minV) = collSmallestIter minV     (pcs':matches) pcss
>     | otherwise          = collSmallestIter minV     matches        pcss

> findMostPacked :: PitchClassSet -> [PitchClassSet] -> PitchClassSet
> findMostPacked pcs' []  = pcs'
> findMostPacked pcs1 (pcs2:pcss)
>   = findMostPacked (mostPacked pcs1 pcs2) pcss

mostPacked determines which set has the most notes close to the first note.

> mostPacked :: PitchClassSet -> PitchClassSet -> PitchClassSet
> mostPacked pcs1 pcs2 = mostPacked' (differenceList pcs1 pcs2) pcs1 pcs2 where
>   mostPacked' []      pcs1 _    = pcs1 -- both sets are the same
>   mostPacked' (val:_) pcs1 pcs2 = if (val > 0) then pcs2 else pcs1
>   differenceList pcs1 pcs2 
>     = dropWhile (== 0) (zipWith (\pc1 pc2 -> pc1 - pc2) pcs1 pcs2)

To rotate a PC Set, simply move the first number to the end and add 12 to it (i.e. shift it up an octave).

> allRotations :: PitchClassSet -> [PitchClassSet]
> allRotations pcs = pcs : (scanAcc rotate' ((length pcs) - 1) pcs) where
>   rotate' (pc:pcs') = pcs' ++ [pc + 12]

*************************************
* Proof of a Property of Prime Forms
*************************************

  "If two PC Sets differ only by transposition or inversion, then they belong to the same Set Class."
  
    - Paul Nelson 

  This is one example of a program propterty that one might like to prove, to assure the rigor of the prime form calculator implementation.  As is mentionned on several Music Theory websites, while there are many prime form calculators freely available, not all of them are correct.  A proof of this property for this program might help assure potential users of its accuracy.

  For the most part, this proof relies on induction and some basic facts about modulo arithmetic.  Near the end, I ran out of steam, so the two (relatively trivial) lemmas required for Theorm 2's proof are left as an exercise for the reader.

  Lemma 1.  (a + b) `mod` c == ((a `mod` c) + (b `mod` c)) `mod` c

    Proof.  Follows from the properties of modulo arithmetic.  

  Lemma 2.  a `mod` b == (a `mod` b) `mod` b

    Proof.  Follows from the properties of modulo arithmetic.  

  Lemma 3. transpose n [] == []
  
  Proof by Calculation.

    transpose n []
    (normalizePCS . (map (+ n))) []
    normalizePCS (map (+ n) [])
    normalizePCS []
    map normalize []
    []

  Lemma 4. normalize ((normalize (a + b)) + (negate (c + b))) == normalize (a - c)

  Proof by Calculation.

    normalize ((normalize (a + b)) + (negate (c + b)))
    ((normalize (a + b)) + (negate (c + b))) `mod` 12 
    (((a + b) `mod` 12) + (negate (c + b))) `mod` 12 
    (((a + b) `mod` 12) - (c + b)) `mod` 12 
    (((a + b) `mod` 12) - c - b) `mod` 12 
    { Lemma 1 }
    ((((a + b) `mod` 12) `mod` 12) - (c `mod` 12) - (b `mod` 12)) `mod` 12 
    { Lemma 2 }
    (((a + b) `mod` 12) - (c `mod` 12) - (b `mod` 12)) `mod` 12 
    { Lemma 1 }
    ((a + b) - c - b) `mod` 12 
    (a + b - c - b) `mod` 12 
    (a + (b - b) - c) `mod` 12 
    (a + 0 - c) `mod` 12 
    (a - c) `mod` 12
    normalize (a - c)

  Lemma 5.  (transpose (negate (a + b))) . (transpose b) == transpose (negate a)

  Proof by Induction.

    Base case: pcs == []

    (transpose (negate (a + b))) . (transpose b) []
    transpose (negate (a + b)) (transpose b [])
    { Lemma 3 }
    transpose (negate (a + b)) []
    { Lemma 3 }
    []
    { Lemma 3 }
    transpose (negate a) []

    Induction step: pcs == pc:pcs', where 
      Induction Hypothesis: ((transpose (negate (a + b))) . (transpose b)) pcs' == transpose (negate a) pcs'

    (transpose (negate (a + b))) . (transpose b) pc:pcs'
    transpose (negate (a + b)) (transpose b pc:pcs')
    transpose (negate (a + b)) ((normalizePCS . (map (+ b))) pc:pcs')
    transpose (negate (a + b)) (normalizePCS (map (+ b) pc:pcs'))
    transpose (negate (a + b)) (normalizePCS (pc + b):(map (+ b) pcs'))
    transpose (negate (a + b)) (map normalize (pc + b):(map (+ b) pcs'))
    transpose (negate (a + b)) (normalize (pc + b)):(map normalize (map (+ b) pcs'))
    transpose (negate (a + b)) (normalize (pc + b)):(normalizePCS (map (+ b) pcs'))
    transpose (negate (a + b)) (normalize (pc + b)):(normalizePCS . (map (+ b)) pcs')
    transpose (negate (a + b)) (normalize (pc + b)):(transpose b pcs')
    (normalizePCS . (map (+ (negate (a + b))))) (normalize (pc + b)):(transpose b pcs')
    normalizePCS (map (+ (negate (a + b))) (normalize (pc + b)):(transpose b pcs'))
    normalizePCS ((normalize (pc + b)) + (negate (a + b))):(map (+ (negate (a + b))) (transpose b pcs'))
    map normalize ((normalize (pc + b)) + (negate (a + b))):(map (+ (negate (a + b))) (transpose b pcs'))
    (normalize ((normalize (pc + b)) + (negate (a + b)))):(map normalize (map (+ (negate (a + b))) (transpose b pcs')))
    (normalize ((normalize (pc + b)) + (negate (a + b)))):(normalizePCS (map (+ (negate (a + b))) (transpose b pcs')))
    (normalize ((normalize (pc + b)) + (negate (a + b)))):(normalizePCS . (map (+ (negate (a + b)))) (transpose b pcs'))
    (normalize ((normalize (pc + b)) + (negate (a + b)))):(transpose (negate (a + b)) (transpose b pcs'))
    { Lemma 4 }
    (normalize (pc - a)):(transpose (negate (a + b)) (transpose b pcs'))
    (normalize (pc - a)):(((transpose (negate (a + b))) . (transpose b)) pcs')
    { Induction Hypothesis }
    (normalize (pc - a)):(transpose (negate a) pcs')
    (normalize (pc - a)):((normalizePCS . (map (+ (negate a)))) pcs')
    (normalize (pc - a)):(normalizePCS (map (+ (negate a)) pcs'))
    (normalize (pc - a)):(map normalize (map (+ (negate a)) pcs'))
    map normalize (pc - a):(map (+ (negate a)) pcs')
    normalizePCS (pc - a):(map (+ (negate a)) pcs')
    normalizePCS ((negate a) +  pc):(map (+ (negate a)) pcs')
    normalizePCS ((+ (negate a)) pc):(map (+ (negate a)) pcs')
    normalizePCS (map (+ (negate a)) pc:pcs')
    (normalizePCS . (map (+ (negate a)))) pc:pcs'
    transpose (negate a) pc:pcs'

  Lemma 6.  transpose n pc:pcs == (pc + n):(transpose n pcs)  
 
  Proof by Calculation.

    transpose n pc:pcs
    map (+ n) pc:pcs
    (pc + n):(map (+ n) pcs)
    (pc + n):(transpose n pcs)

  Lemma 7.  transpose20H (transpose n pcs) == transpose20H pcs

  Proof by Case Analysis.

    case pcs == []:

    transpose20H (transpose n []) 
    { Lemma 3 }
    transpose20H []

    case pcs == pc:pcs'

    transpose20H (transpose n pc:pcs') 
    { Lemma 6 }
    transpose20H (pc + n):(transpose n pcs')
    transpose (negate (pc + n)) (pc + n):(transpose n pcs')
    { Lemma 6 }
    transpose (negate (pc + n)) (transpose n pc:pcs')
    transpose (negate (n + pc)) (transpose n pc:pcs')
    (transpose (negate (n + pc))) . (transpose n) pc:pcs'
    { Lemma 5 }
    transpose (negate pc) pc:pcs'
    transpose20H pc:pcs'

  Lemma 8.  sort (transpose n pcs) == transpose n (sort pcs)

  Proof.  Follows from inequality laws.  if a > b, then (a+n) > (b+n).

  Lemma 9.  normalForm0 (transpose n pcs) == normalForm0 pcs

  Proof by Calculation.

    (Since a large block of code in normalForm0 does not change between calculations in this proof, that code is elided here for clarity)

    normalForm0 (transpose n pcs) 
    ... transpose20H (sort (transpose n pcs)) ...
    { Lemma 8 }
    ... transpose20H (transpose n (sort pcs)) ...
    { Lemma 7 }
    ... transpose20H (sort pcs) ...
    normalForm0 pcs

  Lemma 10.  invertPCS (transpose n pcs) == transpose (negate n) (invertPCS pcs)

  Proof by Induction.

    base case:

    invertPCS (transpose n [])
    { Lemma 3 }
    invertPCS []
    []
    { Lemma 3 }
    transpose (negate n) []
    transpose (negate n) (invertPCS [])

    induction step:   

    invertPCS (transpose n pc:pcs)
    invertPCS ((normalizePCS . (map (+ n))) pc:pcs)
    invertPCS (normalizePCS (map (+ n) pc:pcs))
    invertPCS (normalizePCS (pc+n):(map (+ n) pcs))
    invertPCS (normalize (pc+n)):(normalizePCS (map (+ n) pcs))
    (normalizePCS . (map (12 -))) (normalize (pc+n)):(normalizePCS (map (+ n) pcs))
    normalizePCS (map (12 -) (normalize (pc+n)):(normalizePCS (map (+ n) pcs)))
    normalizePCS (12 - (normalize (pc+n))):(map (12 -) (normalizePCS (map (+ n) pcs)))
    (normalize (12 - (normalize (pc+n)))):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    ((12 - (normalize (pc+n))) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    ((12 - ((pc+n) `mod` 12)) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    { Lemma 1 }
    (((12 `mod` 12) - (((pc+n) `mod` 12) `mod` 12)) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    ((0 - (((pc+n) `mod` 12) `mod` 12)) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    { Lemma 2 }
    ((0 - ((pc+n) `mod` 12)) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    ((((0 `mod` 12) - ((pc+n) `mod` 12)) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    (((0 `mod` 12) - ((pc+n) `mod` 12)) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    { Lemma 1 }
    ((0  - (pc+n)) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    (((-pc) - n) `mod` 12):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    (normalize ((-pc) - n)):(normalizePCS (map (12 -) (normalizePCS (map (+ n) pcs))))
    (normalize ((-pc) - n)):(normalizePCS (map (12 -) (transpose n pcs)
    (normalize ((-pc) - n)):(invertPCS (transpose n pcs))
    { Induction Hypothesis }
    (normalize ((-pc) - n)):(transpose (negate n) (invertPCS pcs))
    (((-pc) - n) `mod` 12):(transpose (negate n) (invertPCS pcs))
    ((0 + ((-pc) - n)) `mod` 12):(transpose (negate n) (invertPCS pcs))
    { Lemma 1 }
    (((0 `mod` 12) + (((-pc) - n) `mod` 12)) `mod` 12):(transpose (negate n) (invertPCS pcs))
    ((0 + (((-pc) - n) `mod` 12)) `mod` 12):(transpose (negate n) (invertPCS pcs))
    (((12 `mod` 12) + (((-pc) - n) `mod` 12)) `mod` 12):(transpose (negate n) (invertPCS pcs))
    { Lemma 1 }
    ((12 + ((-pc) - n)) `mod` 12):(transpose (negate n) (invertPCS pcs))
    (((12 - pc)  - n) `mod` 12):(transpose (negate n) (invertPCS pcs))
    ((((12 - pc) `mod` 12) - (n `mod` 12)) `mod` 12):(transpose (negate n) (invertPCS pcs))
    { Lemma 2 }
    (((((12 - pc) `mod` 12) `mod 12) - (n `mod` 12)) `mod` 12):(transpose (negate n) (invertPCS pcs))
    { Lemma 1 }
    ((((12 - pc) `mod` 12) - n) `mod` 12):(transpose (negate n) (invertPCS pcs))
    (((normalize (12 - pc)) - n) `mod` 12):(transpose (negate n) (invertPCS pcs))
    (normalize ((normalize (12 - pc)) - n)):(transpose (negate n) (invertPCS pcs))
    (normalize ((negate n) + (normalize (12 - pc)))):(transpose (negate n) (invertPCS pcs))
    (normalize ((negate n) + (normalize (12 - pc)))):(normalizePCS (map (+ (negate n)) (invertPCS pcs)))
    normalizePCS ((negate n) + (normalize (12 - pc))):(map (+ (negate n)) (invertPCS pcs))
    normalizePCS (map (+ (negate n)) (normalize (12 - pc)):(invertPCS pcs))
    (normalizePCS . (map (+ interval))) ((negate n) (normalize (12 - pc)):(invertPCS pcs))
    transpose (negate n) (normalize (12 - pc)):(invertPCS pcs)
    transpose (negate n) (normalize (12 - pc)):(normalizePCS (map (12 -) pcs))
    transpose (negate n) (normalizePCS (12 - pc):(map (12 -) pcs))
    transpose (negate n) (normalizePCS (map (12 -) pc:pcs))
    transpose (negate n) ((normalizePCS . (map (12 -))) pc:pcs)
    transpose (negate n) (invertPCS pc:pcs)

  Lemma 11.  (invertPCS (invertPCS pcs)) == pcs, where pcs == normalizePCS pcs

    Left as an exercise for the reader.

  Lemma 12.  mostPacked a b == mostPacked b a 

    Left as an exercise for the reader.
  
  Theorem 1.  primeForm (transpose n pcs) == primeForm pcs

  Proof by Calculation.

    primeForm (transpose n pcs) 
    mostPacked (normalForm0 (transpose n pcs)) (normalForm0 (invertPCS (transpose n pcs)))
    { Lemma 9 }
    mostPacked (normalForm0 pcs) (normalForm0 (invertPCS (transpose n pcs)))
    { Lemma 10 }
    mostPacked (normalForm0 pcs) (normalForm0 (transpose (negate n) (invertPCS pcs)))
    mostPacked (normalForm0 pcs) (normalForm0 (invertPCS pcs))
    primeForm pcs

  Theorem 2.  primeForm (invertPCS pcs) == primeForm pcs, where pcs == normalizePCS pcs

  Proof by Calculation.

    primeForm (invertPCS pcs) 
    mostPacked (normalForm0 (invertPCS pcs)) (normalForm0 (invertPCS (invertPCS pcs))
    { Lemma 11 }
    mostPacked (normalForm0 (invertPCS pcs)) (normalForm0 pcs))
    { Lemma 12 }
    mostPacked (normalForm0 pcs) (normalForm0 (invertPCS pcs))
    primeForm pcs

*******************
* Interval Vectors
*******************

  Interval Vectors -- another staple of atonal analysis.

  "An "Interval Vector" is a list of six numbers which summarizes the interval content in a PC Set. 
  Pitch Intervals
    * The distance between any two pitches is called a "pitch interval". This is the standard definition for an interval in music. For example:
  Interval Classes
    * In the same way that many pitches "sound alike" and are therefore put into the same Pitch Class, there are also many intervals which sound alike and so are put into the same Interval Class.
    * There are six different interval classes which are numbered from 1 to 6.
          o m2 / M7      => 1    (half-steps)
          o M2 / m7      => 2    (whole-steps)
          o m3 / M6      => 3    (minor thirds)
          o M3 / m6      => 4    (major thirds)
          o P4 / P5      => 5    (perfect intervals)
          o A4 / d5      => 6    (tritones)
  Interval Vectors
    * An Interval Vector is a summary of all of the intervals between all pairs of pitches in a pitch class set. It is, essentially, a histogram of all of the interval classes which can be found in a PC Set.
    * For example, a C major chord is represented by the PC Set:  [047]  and has the interval vector <001110>. This is because a C major chord contains one minor third (from E to G) one major third (from C to E) and one perfect interval from C to G). Since a major chord contains no half steps, whole steps, or tritones, these entries in the interval vector are all set to zero (0)."

  - Paul Nelson 

> intervalVector :: PitchClassSet -> IntervalVector
> intervalVector pcs = foldl addInterval initVector allIntervals where
>   allIntervals = listCombinations pcs
>   initVector = array (1,6) [(i, 0) | i <- [1..6]]
>   addInterval :: IntervalVector -> (PitchClassRep, PitchClassRep) -> IntervalVector
>   addInterval vector (pc1, pc2) 
>     | pc1 == pc2 = vector
>     | pc1 >  pc2 = addInterval vector (pc2, pc1) 
>     | otherwise  = vector // [(intervalClass, (1 + (vector ! intervalClass)))] where
>     intervalClass = min (pc2 - pc1) ((pc1 + 12) - pc2)

> intervalVector2List :: IntervalVector -> [Int]
> intervalVector2List iv = [iv ! idx | idx <- [1..6]] 

************
* Rhythm
************

  Rhythm, an often neglected topic, is perhaps the essential aspect of all music.  Interestingly, the increasing focus of 20th c. western art music on pitch has only made rhythm more important.  Specifically, rhythm brings "life" to a pitch class set.  Since a PCS separates a melody from its pitch content, it is useful then to have a data type for the remainder. 

> data Rhythm = NoteR Dur | RestR Dur | Rhythm :&: Rhythm
>     deriving (Show, Eq)    

> numNotes :: Rhythm -> Int
> numNotes (NoteR _) = 1
> numNotes (RestR _) = 0
> numNotes (r1 :&: r2) = (numNotes r1) + (numNotes r2)

  Augmentation: the lengthening of duration of the notes of a phrase
  Diminution:   the inverse of Augmentation

  Haskore has a constructor function for Augmentation, so one implementation of an augmentation function might simply delegate to this constructor.

> augmentWithTempo :: (Ratio Int) -> MusicTransform
> augmentWithTempo = Tempo 

  This approach, however, does not distinguish between tempo modulation and augmentation.  While, technically speaking, there is no difference in interpretation between equivalent tempo modulations and augmentations, it is common convention to use one or the other in various settings, rather than tempo modulation at the exclusion of augmentation.  This practice is certainly redundant when the only interpretation is aural, but when the product of interpretation is a (printed) score, the difference is significant.  Frequent tempo modulations can be hard for the performer to read, especially when they are rendered as successive time signatures.    

  The following implementation of augmentation has the added benefit of more closely matching the usual intent of augmentation which is to change the durations of notes, rather than to change the tempo of a given phrase.

  Note that the use of "augment" is a bit loose here, since this function also allows one to diminish the phrase.

> augment :: (Ratio Int) -> MusicTransform
> augment ratio (Note pitch dur attributes) = Note pitch (dur*ratio) attributes
> augment ratio (Rest dur)                  = Rest (dur*ratio)
> augment ratio (music1 :+: music2)         = (augment ratio music1) :+: (augment ratio music2)
> augment ratio (music1 :=: music2)         = (augment ratio music1) :=: (augment ratio music2)
> augment ratio (Tempo oldRatio music)      = Tempo (oldRatio*ratio) music -- hard to know what to do here, but this seems best
> augment ratio (Trans interval music)      = Trans interval (augment ratio music)
> augment ratio (Instr iName music)         = Instr iName (augment ratio music)
> augment ratio (Player pName music)        = Player pName (augment ratio music)
> augment ratio (Phrase attributes music)   = Phrase attributes (augment ratio music)

  Masking Music

  In a certain sense, large-scale forms are also a type of rhythm.  The following is a simple operation on forms that I have not yet seen in other composition software.  The basic idea is to apply the concept of "masks" or "bit masks" from image processing, to music.  The general idea is that the mask is a pattern which can used to select regions of the media.  Then a specific function is applied to the regions that are selected, while the other regions are left alone.  In the following, the function "cut" is used, but it is certainly possible to implement this concept, using any other music manipulation function.

> type SequenceMask = [Bool]

  subtractPCS takes a piece of music and subtracts some of its notes, based on a "mask" where True marks a note we keep, and False notes a note we disregard

> subtractPCS :: SequenceMask -> PitchClassSetTransform
> subtractPCS []         pcs      = pcs
> subtractPCS _          []       = []
> subtractPCS (True:ms)  (pc:pcs) = pc:(subtractPCS ms pcs)
> subtractPCS (False:ms) (_:pcs)  = subtractPCS ms pcs

> subtractM :: SequenceMask -> MusicTransform
> subtractM []        _                 = error "Sequence Mask was not long enough to process entire music value"
> subtractM (True:_)  note@(Note _ _ _) = note
> subtractM (False:_) note@(Note _ _ _) = emptyMusic -- no convenient way to represent an "empty" music value
> subtractM (True:_)  rest@(Rest _)     = rest
> subtractM (False:_) rest@(Rest _)     = emptyMusic
> subtractM mask (music1 :+: music2)    = (subtractM mask1 music1) :+: (subtractM mask2 music2) where
>   numSub1 = (numSubtractable music1)
>   mask1   = take numSub1 mask
>   mask2   = drop numSub1 mask
> subtractM mask (music1 :=: music2)    = (subtractM mask music1) :=: (subtractM mask music2)
> subtractM mask (Tempo ratio music)    = Tempo ratio (subtractM mask music)
> subtractM mask (Trans interval music) = Trans interval (subtractM mask music)
> subtractM mask (Instr iName music)    = Instr iName (subtractM mask music)
> subtractM mask (Player pName music)   = Player pName (subtractM mask music)
> subtractM mask (Phrase att music)     = Phrase att (subtractM mask music) 

A simple helper function to allow the "bitmasking" of music values that use :+:.

> numSubtractable :: (Ord a, Num a) => Music -> a
> numSubtractable (Note _ _ _)        = 1
> numSubtractable (Rest _)            = 1
> numSubtractable (music1 :+: music2) = (numSubtractable music1) + (numSubtractable music2)
> numSubtractable (music1 :=: music2) = max (numSubtractable music1) (numSubtractable music2)
> numSubtractable (Tempo _ music)     = numSubtractable music
> numSubtractable (Trans _ music)     = numSubtractable music
> numSubtractable (Instr _ music)     = numSubtractable music
> numSubtractable (Player _ music)    = numSubtractable music
> numSubtractable (Phrase _ music)    = numSubtractable music

  Note Density

  noteDensity measures the number of articulations of distinct notes per unit of time, where the unit of time is the same one employed by the "dur" function.  This effectively returns something comparable to a notes-per-bar measurment, adjusted for tempo scaling.  Such measures can be quite useful in music analysis, but are tedious to perform by hand.

> noteDensity :: Music -> Ratio Int
> noteDensity music = (fromInt (numArticulations music)) / (dur music)

> numArticulations :: Music -> Int
> numArticulations (Note _ _ _)        = 1
> numArticulations (Rest _)            = 0
> numArticulations (music1 :+: music2) = (numArticulations music1) + (numArticulations music2)
> numArticulations (music1 :=: music2) = (numArticulations music1) + (numArticulations music2)
> numArticulations (Tempo _ music)     = numArticulations music
> numArticulations (Trans _ music)     = numArticulations music
> numArticulations (Instr _ music)     = numArticulations music
> numArticulations (Player _ music)    = numArticulations music
> numArticulations (Phrase _ music)    = numArticulations music

*************
* An Example
*************

  To demonstrate the use of these extensions to Haskore, Bartok's Mikrokosmos #37 is encoded, analyzed, and a set of (micro-)variations based on it is generated.

  Analysis

  Note the micro-theme in the following analaysis via haskore.  Interestingly this theme could easily be detected by a standard search algorithm.  It also also has a characteristic rhythmic figure, which might be reduced to "long short short long".  Of course, the rhythmic notation used here is not yet flexible enough to model these abstractions fully.  

  Also note the round-like repetition of the main theme, and the way that the repitition comes back, in the middle, more quickly, using the micro-theme.  Repitition also occurs at the end, where the left hand literally repeats itself.  I believe that the Haskell encoding of the score brings out these facets nicely in the decomposition of pitch classes.  This, I believe, suggests the suitability of Haskore as an executable language for musical analysis.  

> mikrokosmos37 :: Music  
> mikrokosmos37 = Tempo 2 (Trans 5 (rightHandPart :=: leftHandPart)) where
>  rightHandPart :: Music
>  rightHandPart  = makeMelody (5*12) rightPCS rightRhythm
>  rightRhythm    = (RestR 2) :&: dhn :&: qtn :&: qtn :&: e2 :&: q2 :&: q3e2 :&:
>	            hn :&: qr :&: q2 :&: q2 :&: e2 :&: e4 :&: q2 :&: qr :&: hn :&: hn :&:
>	            q2 :&: e2 :&: dhn :&: e2 :&: e4 :&: e4 :&: (NoteR (5 % 8)) :&: en :&:
>	            e2 :&: e4 :&: e4 :&: qr :&: e2 :&: q2 :&: qtn :&: e2
>  rightPCS :: PitchClassSet
>  rightPCS       = introPhraseP ++ bellPhraseP ++ bellPhrase2P ++ 
>	            decorPhraseP ++ decorPhrase2P ++ resoPhraseP
>  bellPhraseP    = (replicate 4 7) ++ [6,2,4,2,0,2,4,0]
>  bellPhrase2P   = (replicate 3 7) ++ microThemeP
>  decorPhraseP   = [4,6,7,4,6,2,4,2,4,0,2]
>  decorPhrase2P  = [0,4,6,7,6,4,6,4,2,0,2]
>  resoPhraseP    = [4,6,4,2,0,2,0]
>  microThemeP    = [6,4,6,2]
>  leftHandPart :: Music
>  leftHandPart   = makeMelody (4*12) leftPCS leftRhythm 
>  leftRhythm     = introPhraseR :&: phraseL2R :&: joinPhraseR :&: steadyPhraseR :&: 
>	            steady2PhraseR :&: angularPhraseR :&: risingPhraseR :&: risingPhrase2R
>  introPhraseR   = q3e2 :&: q2 :&: q3e2 :&: dhn
>  phraseL2R      = q2 :&: e2 :&: q2 :&: dhn
>  joinPhraseR    = e2 :&: qtn
>  steadyPhraseR  = hn :&: hn :&: hn :&: e2 :&: e4
>  steady2PhraseR = q2 :&: q2 :&: q2
>  angularPhraseR = qr :&: q2 :&: e2 :&: en :&: er :&:hn :&: qtn :&: (NoteR (5 % 8)) :&: 
>	            e2 :&: en :&: dhn
>  risingPhraseR  = e2 :&: q2 :&: dhn
>  risingPhrase2R = e2 :&: q2 :&: (NoteR 1)
>  leftPCS        = introPhraseP ++ phraseL2P ++ joinPhraseP ++ 
>	            steadyPhraseP ++ steady2PhraseP ++ angularPhraseP ++ 
>	            risingPhraseP ++ risingPhraseP
>  introPhraseP   = [0,7] ++ microThemeP ++ [0,4,7] ++ microThemeP
>  phraseL2P      = [7,6,4,2,0,4,2]
>  joinPhraseP    = [4,6,2]
>  steadyPhraseP  = (replicate 3 7) ++ [6,0,7,6,4,6]
>  steady2PhraseP = replicate 6 7
>  angularPhraseP = [7] ++ microThemeP ++ [7,0,6] ++ [4,6,2] ++ [4]
>  risingPhraseP  = [0,2] ++ [4,6,7]
>  q2             = qtn :&: qtn
>  e2             = en :&: en
>  e4             = e2 :&: e2
>  q3e2           = q2 :&: qtn :&: e2
>  qtn            = NoteR (1 % 4)
>  en             = NoteR (1 % 8)
>  hn             = NoteR (1 % 2)
>  dhn            = NoteR (3 % 4)
>  qr             = RestR (1 % 4)
>  er             = RestR (1 % 8)

  Synthesis

  Now for the interesting part.  Using that (rather simple) piece as our source material, we will construct some small snippets of origional music, using the tools in the workbench.  This will serve both as a demonstration of the use of these functions.  It also, demonstrates the way in which programming languages can bring together the analytic and synthetic aspects of music...

  First, a demo of analysis functions.

  The heading for Mikrokosmos #37 is "In Lydian Mode", and, in fact, the entire piece is restricted entirely to the first 5 notes of the lydian mode.

> lydianMode :: PitchClassSet
> lydianMode = [0,2,4,6,7,9,11]

> m37PCS :: PitchClassSet
> m37PCS = take 5 lydianMode

> lydianPF :: PitchClassSet
> lydianPF = primeForm lydianMode -- [0,1,3,5,6,8,10]

> m37PF :: PitchClassSet
> m37PF = primeForm m37PCS -- [0,1,3,5,7]

  Note that, since it has the same interval content as the major scale, the lydian (or 4th) mode of the major scale has the same prime form as every other mode of the major scale.

> majorMode :: PitchClassSet
> majorMode = [0,2,4,5,7,9,11]

> majorModePF :: PitchClassSet
> majorModePF = primeForm majorMode -- [0,1,3,5,6,8,10]

  So what about interval content?  Again, we are in the realm of the major scale here (with it's infameous wwhwwwh construction).

> lydianIV :: [Int]
> lydianIV = intervalVector2List $ intervalVector lydianMode -- <2,5,4,3,6,1>

> majorModeIV :: [Int]
> majorModeIV = intervalVector2List $ intervalVector majorMode -- <2,5,4,3,6,1>

> m37IV :: [Int]
> m37IV = intervalVector2List $ intervalVector m37PCS -- <1,3,1,2,2,1>

  So Bartok manages to select a subset of the mode that contains at least one of every type of interval.  It also seems to be "weighted" towards major seconds, major thirds, and fourths.  I think that the overall "sound" of this piece reflects this. 

  Now for some transformations...

  Lets make a new set of pitches by transposing the micro-theme identified earlier through all of its component interval vectors.  (A pure contrivance, I know, but maybe it will sound good...)

> m37MicroTheme :: PitchClassSet
> m37MicroTheme = [6,4,6,2]

> m37MicroThemeIV = intervalVector2List $ intervalVector m37MicroTheme -- <030200>

> m37MTSet = m37MicroTheme ++ (transposeB 2 m37MicroTheme) ++ (transposeB 4 m37MicroTheme) -- [6,4,6,2,8,6,8,4,10,8,10,6]

  So how does it sound?

> m37MTSetM = pcs2Melody m37MTSet

> playM37MTSet :: IO ()
> playM37MTSet = testMac (Tempo 4 m37MTSetM) "m37MTSet"

  Well, That's alright, but what about mutating it a bit more?

> m37MTSCycled = (cycleIn True m37MTSet) ++ (cycleIn False m37MTSet)
> m37MTSCycledM = pcs2Melody m37MTSCycled

> playM37MTSet2 :: IO ()
> playM37MTSet2 = testMac (Tempo 12 m37MTSCycledM) "m37MTSet2"

  To my ear, that experiment didn't work very well ;)

  Here's an example of inversion, and transposition + inversion.

> angularPhrase :: PitchClassSet
> angularPhrase = [7] ++ m37MicroTheme ++ [7,0,6] ++ [4,6,2] ++ [4]

> angPhrase :: PitchClassSet
> angPhrase = (invertPCS angularPhrase) ++ (transposeInvert 7 angularPhrase) ++ angularPhrase

> angPhraseM = pcs2Melody angPhrase

> playAngPhrase :: IO ()
> playAngPhrase = testMac (Tempo 10 angPhraseM) "angPhrase"

  I like that one.  It is dissonant, and its shifting motion keeps the ear alert  :)

  Now let's see if we can't use the multiply function to dress up the PCS that didn't sound good to me.

> mtSet2RevampM :: Music
> mtSet2RevampM = (((Trans (-12) m37MTSCycledM) :=: (delayM 10 2)) :=: (Trans 12 (delayM 24 3))) :=: (Trans 24 (delayM 30 5)) where
>   delayM rDur factor = (Rest rDur) :+: (pcs2Melody (multiplyPCS factor m37MTSCycled))

> playMTS2Rev :: IO ()
> playMTS2Rev = testMac (Tempo 10 mtSet2RevampM) "mtSet2Rev"

  Well, that sounds much better.  Perhaps this further transformation works because the contour is preserved somewhat throughout the large-scale repitions.  
  Let's play around with contour a bit more...

> angPhraseC :: PitchClassSet
> angPhraseC = (contractIntervals (1%4) angPhrase) ++ (contractIntervals (1%2) angPhrase) ++ 
>              angPhrase ++ (contractIntervals 2 angPhrase) ++ 
>	       (contractIntervals 4 angPhrase) ++ (contractIntervals 8 angPhrase)

> angPhraseCM = pcs2Melody angPhraseC

> playAngPhraseC :: IO ()
> playAngPhraseC = testMac (Tempo 10 angPhraseCM) "angPhraseCM"

  This is my favorite snippet.  I feel that it is possible to hear the phrase getting blown up as if it was zoomed in under a musical microscope.  This is the kind of sonic geometry that I think a FP DSL like Haskore can model well.

  So what next?  Bartok uses slight modifications of phrases in imititation.  How about interpolating them at inversion?

> bellDecorInt :: PitchClassSet
> bellDecorInt = (intInv bellPhraseP bellPhrase2P) ++ (intInv decorPhraseP decorPhrase2P) where
>   intInv pcs1 pcs2 = (interpolate pcs1 (invertPCS pcs2))
>   bellPhraseP    = (replicate 4 7) ++ [6,2,4,2,0,2,4,0]
>   bellPhrase2P   = (replicate 3 7) ++ m37MicroTheme
>   decorPhraseP   = [4,6,7,4,6,2,4,2,4,0,2]
>   decorPhrase2P  = [0,4,6,7,6,4,6,4,2,0,2]

> bellDecorIntM = pcs2Melody bellDecorInt

> playBDI :: IO ()
> playBDI = testMac (Tempo 11 bellDecorIntM) "bellDecorInt"

  This one is a bit of a dud.  Oh well, they can't all be keepers!  It's hard to hear the interpolation in the sequence.

  How about taking a phrase we worked on earlier, and realize it as chords instead of melodies?

> angCChords = foldr1 (:+:) (map (:+: (Rest (1%2))) (map pcs2Chord (chunk 5 angPhraseC)))

> playAngCChords = testMac (Tempo 5 angCChords) "angCChords"

  There are some nice chords in there, but the flow of these chords in sequence doesn't really do much.  When played at a decent tempo, I do like the last 20 seconds or so where the chords open up.

  Now for rhythm, let's try dressing up m37MTSCycled again (making lemonade is fun).

> m37MTSReCycled = makeMelody (5*12) m37MTSCycled aRhythm where
>   aRhythm = qtn :&: e2 :&: q2 :&: dhn :&: qtn :&: e3er :&: e3er :&: qr :&: q2 :&: qtn :&: qtn :&: hn :&: en :&: e2 :&: dhn :&: (RestR 1)
>   e3er           = e2 :&: en :&: er
>   q2             = qtn :&: qtn
>   e2             = en :&: en
>   e4             = e2 :&: e2
>   q3e2           = q2 :&: qtn :&: e2
>   qtn            = NoteR (1 % 4)
>   en             = NoteR (1 % 8)
>   hn             = NoteR (1 % 2)
>   dhn            = NoteR (3 % 4)
>   qr             = RestR (1 % 4)
>   er             = RestR (1 % 8)

> mtsReCM = (ver (1%5)) :+: (Trans (-2) (ver (1%4))) :+: (Trans (-5) (ver (1%3))) :+: (Trans (-12) (ver (1%2))) where
>  ver factor = augment factor m37MTSReCycled

> playMTSReC = testMac mtsReCM "mtsReC"

  Finally, here's an example of "BitMasking".

> risingM = melody :+: melody where
>  melody        = makeMelody (5*12) risingPhraseP aRhythm 
>  risingPhraseP = [0,2] ++ [4,6,7]
>  aRhythm       = e2 :&: q2 :&: dhn
>  q2            = qtn :&: qtn
>  e2            = en :&: en
>  qtn           = NoteR (1 % 4)
>  en            = NoteR (1 % 8)
>  dhn           = NoteR (3 % 4)

> risingMSub = risingM :+: m1 :+: m2 :+: m3 :+: m4 where
>   m1 = sub [1,1,0,0,1,0,0,1,1,1]
>   m2 = sub [1,1,1,0,0,0,1,1,1,0]
>   m3 = sub [1,0,1,0,1,0,1,0,1,0]
>   m4 = sub [1,1,0,0,0,0,0,0,1,1]
>   sub ints = subtractM [odd x | x <- ints] risingM

> playRisingSub = testMac (Tempo 2 risingMSub) "risingSub"

  This one is very satisfying for me.  I like being able to look at the masks while the music is playing.  I feel that I can  hear the missing notes...

  Finally, let's put them all together to make a theme and variations on Bartok's Mikrokosmos #37.  

> m37TAndV :: Music
> m37TAndV = themes :+: pause :+: var1 :+: var2 :+: pause :+: var3 :+: pause :+: origional where
>   pause     = (Rest 1)
>   themes    = Tempo 2 (makeMelody (5*12) (m37MicroTheme ++ angularPhrase) themesR)
>   var1      = Tempo 10 (Trans 5 (mtSet2RevampM :+: (retro2 mtSet2RevampM)))
>   var2      = (Tempo 11 (angPhraseCM :+: (Note var2EndP 3 []) :+: 
>                          (Note var2EndP 16 []) :+: (Note var2EndP 16 [])))
>   var3      = ((Tempo 3 (Trans 1 (risingMSub :+: risingMSub))) :=: 
>                (Tempo 2 ((Rest 2) :+: angCC2)))  
>   origional = mikrokosmos37
>   themesR   = qtn :&: e2 :&: qtn :&: angularR
>   angularR  = qr :&: q2 :&: e2 :&: en :&: er :&:hn :&: qtn :&: (NoteR (5 % 8)) :&: 
>	        e2 :&: en :&: dhn
>   var2EndP  = (pitch 25)
>   angCC2    = foldr1 (:+:) (map (:+: (Rest (1%3))) (map pcs2Chord (take 13 (drop 30 chunkAPC))))
>   chunkAPC  = chunk 5 angPhraseC
>   q2        = qtn :&: qtn
>   e2        = en :&: en
>   qtn       = NoteR (1 % 4)
>   en        = NoteR (1 % 8)
>   hn        = NoteR (1 % 2)
>   qr        = RestR (1 % 4)
>   er        = RestR (1 % 8)
>   dhn       = NoteR (3 % 4)

> retro2 :: MusicTransform
> retro2 (m1 :=: m2)   = (retro2 m1) :=: (retro2 m2)
> retro2 (Trans int m) = (Trans int (retro2 m))
> retro2 m             = retro m

> playM37TAndV = testMac m37TAndV "m37ThemeAndVariations"

************************
* Mapping Text to Pitch
************************

  The following was part of my assignment #9, but I thought it would be nice to have all of my Haskore stuff in one place.  (It has been modified to use the latest version of Haskore) It also demonstrates one direction in which I think Haskore could be fruitfully expanded.

  A piece which uses a mapping between the words in Mary Had a Little Lamb and their associated melodic fragments to generate music based on the words in a simple poem that I wrote, also constrained to the words in the lyrics to Mary had a Little Lamb.  I think it sounds alright, considering that I was limited to the lyrical and musical vocabulary of Mary Had a Little Lamb (only three pitches, and three note durations to work with!).  The full lyrics to Mary Had a Little Lamb are courtesy of Google ;)

> poemMusic2 :: Music
> poemMusic2 = poemM2 "Hammond Organ" 12

Mary had a little lamb,
Little lamb, little lamb,
Mary had a little lamb,
Its fleece was white as snow

And everywhere that Mary went,
Mary went, Mary went,
Everywhere that Mary went
The lamb was sure to go

It followed her to school one day
School one day, school one day
It followed her to school one day
Which was against the rules.

It made the children laugh and play,
Laugh and play, laugh and play,
It made the children laugh and play
To see a lamb at school

And so the teacher turned it out,
Turned it out, turned it out,
And so the teacher turned it out,
But still it lingered near

And waited patiently about,
Patiently about, patiently about,
And waited patiently about
Till Mary did appear

"Why does the lamb love Mary so?"
Love Mary so? Love Mary so?
"Why does the lamb love Mary so?"
The eager children cry

"Why, Mary loves the lamb, you know."
Loves the lamb, you know, loves the lamb, you know
"Why, Mary loves the lamb, you know."
The teacher did reply 

> type Text = [String]

> textToMusic :: Text -> Music 
> textToMusic []          = Rest 0
> textToMusic (word:text) = wordToMusic :+: textToMusic text where
>   oct         = 5
>   note1       = c oct qn []
>   note1hn     = c oct hn []
>   note1whn    = c oct wn []
>   note2       = d oct qn []
>   note3       = e oct qn []
>   note3hn     = e oct hn []
>   wordToMusic = case word of
>     "a"          -> note2
>     "about"      -> note3 :+: note3
>     "against"    -> note2 :+: note3 
>     "and"        -> note2 
>     "appear"     -> note2 :+: note1whn 
>     "as"         -> note2
>     "at"         -> note2
>     "but"        -> note3
>     "children"   -> note1 :+: note2 
>     "cry"        -> note1whn
>     "day"        -> note3hn 
>     "did"        -> note3
>     "does"       -> note3
>     "eager"      -> note2 :+: note2
>     "everywhere" -> note3 :+: note2 :+: note1
>     "fleece"     -> note2
>     "followed"   -> note3 :+: note2
>     "go"         -> note1whn
>     "had"        -> note1
>     "her"        -> note1
>     "it"         -> note2
>     "its"        -> note3
>     "know"       -> note3hn
>     "lamb"       -> note3hn
>     "laugh"      -> note3
>     "lingered"   -> note3 :+: note2
>     "little"     -> note3 :+: note3
>     "love"       -> note2
>     "loves"      -> note1
>     "made"       -> note3
>     "Mary"       -> note3 :+: note2
>     "near"       -> note1whn
>     "one"        -> note3
>     "out"        -> note3hn
>     "patiently"  -> note1 :+: note2 :+: note3 
>     "play"       -> note3hn
>     "reply"      -> note1whn
>     "rules"      -> note1whn
>     "school"     -> note3 
>     "see"        -> note2
>     "snow"       -> note1whn 
>     "so"         -> note3
>     "sure"       -> note3
>     "still"      -> note2
>     "teacher"    -> note1 :+: note2
>     "that"       -> note2
>     "the"        -> note3
>     "till"       -> note3
>     "to"         -> note2
>     "turned"     -> note3
>     "waited"     -> note3 :+: note2
>     "was"        -> note2
>     "went"       -> note3hn
>     "which"      -> note3
>     "white"      -> note3
>     "why"        -> note2
>     "you"        -> note3
>     unmatched    -> error ("no music mapping for: " ++ unmatched)   

A Simple Poem Based on the Words to Mary Had a Little Lamb

you went to school against the snow
to know a little to laugh a little
and against the rules you go 
to laugh a little to play a little
and patiently as you do so
the day does go

the snow the snow the white white snow ...

snow everywhere snow everywhere ...

why did you linger so
her laugh was her sure reply
why
It was to love to know 
and you know to see the snow

> poemLine1 :: Text
> poemLine1  = words "you went to school against the snow"
> poemLine2  = words "to know a little to laugh a little"
> poemLine3  = words "and against the rules you did go"
> poemLine4  = words "to laugh a little to play a little"
> poemLine5  = words "and patiently as you went so"
> poemLine6  = words "the day did go"
> poemLine7  = words "the snow the snow the white white snow"
> poemLine8  = words "snow everywhere snow everywhere"
> poemLine9  = words "why had you lingered so"
> poemLine10 = words "her laugh was her sure reply"
> poemLine11 = words "why"
> poemLine12 = words "it was to love to know"
> poemLine13 = words "and you know to see the snow"

> stanza1Music = textToMusic $ poemLine1 ++ poemLine2 ++ poemLine3 ++ 
>                              poemLine4 ++ poemLine5 ++ poemLine6

> stanza2Music = textToMusic $ poemLine9 ++ 
>			       poemLine10 ++ poemLine11 ++ poemLine12 ++ 
>			       poemLine13

> poemM2 inst dtp = (Tempo (7%6) song) where
>   song   = line8Bass :+: middle :+: ending
>   middle = (part1 :+: (Trans 2 part2) :+: part3) `overR` rhythm
>   ending = (endDescant :=: endRhythm) `overR` line7Bass
>   part1  = (addStyle stanza1Music) `overR` line7Bass
>   part2                   
>     = ((addStyle stanza2Music) `overR` ((Rest (1%2)) :+: descant1))
>       `overR`
>       line8Bass
>   part3                   
>     = ((addStyle (stanza2Music :+: (textToMusic ["snow"])))
>	 `overR`
>  	 ((Rest (1%2)) :+: descant2))
>       `overR`
>       (revM line7Bass)
>   endDescant = (descant2 :+: descant1 :+: descant1)
>   endRhythm 
>     =  ((rhythm :+: rhythm :+: rhythm :+: rhythm :+: (Rest 4)))
>   line7Bass  = makeBass (textToMusic poemLine7)
>   line8Bass  = makeBass (textToMusic poemLine8)
>   makeBass m = Trans (-24) (double m "Acoustic Bass" "Hammond Organ")
>   rhythm                  
>     = Tempo 3 (Instr "Percussion" (bd :+: hat :+: (bd :=: tom) :+: 
>			            (hat :=: cowb) :+: bd))
>   bd       = perc BassDrum1 qn []
>   tom      = perc LowMidTom qn []
>   hat      = perc ClosedHiHat en []
>   cowb     = perc Cowbell en []
>   descant2 = makeDescant poemLine6 0
>   descant1 = makeDescant poemLine9 2
>   makeDescant line rest = addDStyle (makePalindrome (textToMusic line) rest)
>   addDStyle m           = (Instr inst (Tempo 6 (Trans dtp m)))
>   makePalindrome m rest = m :+: (Rest rest) :+: revM m
>   addStyle m            = Tempo 3 (double m "Vibraphone" "Ocarina")
>   stat `overR`  dyn     = stat :=: (cut (dur stat) (repeatM dyn)) 
>   repeatM  m            = m :+: repeatM m
>   double m inst1 inst2  = (Instr inst1 m) :=: (Instr inst2 m)

> poemMusic1 :: Music
> poemMusic1 = textToMusic $ poemLine1 ++ poemLine2 ++ poemLine3 ++ 
>                            poemLine4 ++ poemLine5 ++ poemLine6 ++ 
>			     poemLine7 ++ poemLine8 ++ poemLine9 ++ 
>			     poemLine10 ++ poemLine11 ++ poemLine12 ++ 
>			     poemLine13

> poemLine1Music :: Music
> poemLine1Music = textToMusic poemLine1

> playPoemLine1 :: IO ()
> playPoemLine1 = testLin (Tempo 2 poemLine1Music) "mary-poem-line-1"

> playPoem1 :: IO ()
> playPoem1 = testLin (Tempo 3 poemMusic1) "mary-poem-1"

> playPoem1M :: IO ()
> playPoem1M = testMac (Tempo 5 (Instr "Hammond Organ" poemMusic1)) "mary-poem-1"

> playPoem2M :: IO ()
> playPoem2M = testMac poemMusic2 "mary-poem-2"

> testPoem2M inst dtp = testMac (poemM2 inst dtp) "mary-poem-2n"

********************
* Generic Functions
********************

mostly for debugging purposes...

> intVecAndPrimeForm :: PitchClassSet -> (IntervalVector, PitchClassSet)
> intVecAndPrimeForm pcs = (intervalVector pcs, primeForm pcs)  

Exploration Functions ...

> ivAndPF pcs = (intervalVector2List (fst $ ivPF), snd $ ivPF) where
>   ivPF = intVecAndPrimeForm pcs

> test1 :: Music -> String -> IO ()
> test1 m fName = outputMidiFile (fName ++ ".mid") (performToMidi (perform defPMap defCon m) defUpm) 

> testMac :: Music -> String -> IO ()
> testMac m fName = do test1 m fName
>                      system ("open " ++ fName ++ ".mid")
>                      return ()

> testLin :: Music -> String -> IO ()
> testLin m fName = do test1 m fName
>                      system ("timidity " ++ fName ++ ".mid")
>                      return ()

********************
* Future Directions
********************

  My experience with Haskore to-date suggests to me three main (complementary) directions for further Haskore development:  Development of a representation of sound suitable for sound sequencing, development of a representation of scores, suitable for exact transcription and generation of scores, and the development of a set of abstractions for the music theoretic principles which apply to these representations.  This separation of musical sounds from musical instructions is necessary since there is simply not enough knowledge about the details of musical (score) interpretation to accurately formalize the transformation of scores (as traditionally defined) into sound.  Simply put, it will be a long time before my soundcard can play a Bach violin concerto as well as the lead chair at the NY Phil. can, and it may never be able to do so.  Although an interesting topic, I will leave a further discussion of this for another day.  The bottom line is that Haskore currently conflates sound representation with score representation.  Separating these abstractions somewhat will open up many new possibilities for the system.

  Primarily, this development would facilitate the use of Haskore as a formal language for music analysis.  In turn, this capability would allow composers to utilize, easily and directly, the benefits of modern musical analysis, by using analytical algorithms to synthesize new music.  Also, such a language would allow one to create a musical computer -- a processor for both sound and scores that would extract, via controlled automation, the structure of music.

****************
* References
****************

Paul Nelson, http://composertools.com/Theory/
Gary R. Tucker, http://www.mta.ca/faculty/arts/music/course_materials/pc-set/packed.html
Larry J. Solomon, http://mus.comp.home.att.net/
Paul Hudak, "An Algebraic Theory of Polymorphic Temporal Media"
            "School of Expression"
            "Haskore Tutorial"
Bela Bartok, Mikrokosmos #37

==============================================================================================================================



