\section{Performance Evaluation}

To build a database of information about a performer's abilities one must provide a means to input this information in a machine processible format.  While it would be ideal to extract this information from the recorded audio and video of a set of performances, such a task is beyond the scope of this research project.  On the other extreme, requiring the performer to measure in detail his abilities and enter them into a database is impractical.  The {\tt PerformanceEvaluation} module attempts to strike a practical balance between ease of use and ease of implementation in this regard.

\begin{verbatim}

> module PerformanceEvaluation where
> import Basics
> import SaxophoneFingerings
> import Performer
> import Scores

\end{verbatim}

The {\tt PerformanceEvaluation} data type represents an evaluation of a given performance.  This evaluation is of a particular phrase in a larger work, and it indicates whether or not the performer played that phrase successfully.  {\tt PerformanceEvaluation}s are linked to the phrases they describe by a phrase identifier, which is a score identifier (the title of the score) combined with an integer which indicates the phrase number (0 for the first phrase, 1 for the second, etc.) of that score.  If the performer did not play the phrase successfully, a reason is provided.

\begin{verbatim}

> data PerformanceEvaluation 
>  = CanPlay PhraseID | CannotPlay PhraseID Reason
>     deriving (Read, Show, Eq)

> data PhraseID = PhraseNum Integer ScoreID
>     deriving (Read, Show, Eq)

> type ScoreID = String 

> data Reason = TooHigh | TooLow | TrickyRhythm 
>             | TooLong | TrickyFingering
>     deriving (Read, Show, Eq)

\end{verbatim}

  This representation is modelled after the common practice of marking up a score with annotations.  For example, a music teacher might read the score while the student plays, and use a pencil to circle and annotate phrases that were played incorrectly.  By encoding those annotations in a machine processible form, a foundation is created for building a rough measure of a performer's technical facility, programmatically.  The system may then develop further etudes that will help ``hone in'' on the performer's weaknesses in technique.

\subsection{Updating the Performer Database}

  Performance evaluations include useful information about the performer's technical facility that one would like to extract.  For each performance, it would be helpful to find the phrases that were successful or failed due to difficulty of technique, and assimilate that information into the performer database.  The {\tt updateWithEvals} function takes a list of scores, a list of performance evaluations, and a performer database, and it updates that database according to information derived from the performance evaluations.

\begin{verbatim}

> updateWithEvals :: [Score] -> [PerformanceEvaluation] -> 
>                    TechniqueMap -> TechniqueMap
> updateWithEvals s [] t = t
> updateWithEvals s (p:ps) t 
>   = updateWithEvals s ps (updateWithEval s p t)

> updateWithEval :: [Score] -> PerformanceEvaluation -> 
>                   TechniqueMap -> TechniqueMap
> updateWithEval s (CanPlay p) t 
>   = updateWithEvalBy updateWithSuccess s p t
> updateWithEval s (CannotPlay p TrickyFingering) t 
>   = updateWithEvalBy updateWithFailure s p t
> -- no assimilations yet for other types of failure
> updateWithEval _ (CannotPlay _ _) t = t 

> updateWithEvalBy :: (Music -> Integer -> Integer -> 
>                      TechniqueMap -> TechniqueMap) -> 
>                     [Score] -> PhraseID -> 
>                     TechniqueMap -> TechniqueMap
> updateWithEvalBy f scores (PhraseNum i title) t = f m bpm bpMeas t
>   where m :: Music 
>         m      = phraseNum i s
>         bpMeas = timeSigToBPMeas timeS
>         s@(Score _ timeS bpm _ _) = scoreWithTitle title scores
 
\end{verbatim}

  This function is implemented with two update functions -- one which is applied for successful phrases, and another that is applied for phrases that were performed unsuccessfully.  In the case of a phrase that was successfully performed, the database is updated to reflect the performer's ability to play each of the fingering transitions that occurred within the phrase.  The {\tt updateWithSuccess} function takes a phrase, a tempo in beats per minute, and a number of beats per measure, and returns a function which updates the performer database accordingly.

  For each transition in the phrase, if the speed of that transition in the phrase is higher than the maximum speed of that transition in the database, then the database is updated with the new speed.  If the speeds are equal, then the confidence in that transition is incremented.  Otherwise the confidence in that interval is incremented by a small amount, but not more than a small amount over the ``low confidence'' level.

\begin{verbatim}

> updateWithSuccess :: Music -> Integer -> Integer -> 
>                      TechniqueMap -> TechniqueMap
> updateWithSuccess = updateBy updateTSucc 

> updateTSucc :: (SaxFCombination, FingeringTempo) -> 
>                TechniqueMap -> TechniqueMap
> updateTSucc (d,pt) t' = updateRecord t' d newT newC
>   where kbt  = maximumTempo t' d
>         kbc  = confidence t' d
>         newT = if kbt < pt then pt else kbt
>         newC = if kbt == pt 
>                then min (kbc + 1) maxConfidence 
>                else (if kbc < (lowConfidence + 1) then kbc + 1 else kbc)

\end{verbatim}

  For phrases that were unsuccessfully performed due to technique problems, it is not certain which of the transitions within the phrase are the source of difficulty, so every transitions' confidence level is decreased.  In the etude generation step, these ``low confidence'' transitions will be selected for testing so as to ``hone in'' on the particularly difficult transitions for that performer.  Like {\tt updateWithSuccess}, {\tt updateWithFailure} takes a phrase, a tempo in beats per minute, and a number of beats per measure, and returns a function which updates the performer database accordingly.

\begin{verbatim}

> updateWithFailure :: Music -> Integer -> Integer -> 
>                      TechniqueMap -> TechniqueMap
> updateWithFailure = updateBy updateTFail 

> updateTFail :: (SaxFCombination, FingeringTempo) -> 
>                TechniqueMap -> TechniqueMap
> updateTFail (d,pt) t' = updateRecord t' d kbt newC
>   where kbt  = maximumTempo t' d
>         kbc  = confidence t' d
>         newC = if kbc > lowConfidence 
>                then lowConfidence 
>                else max minConfidence (kbc - 1)

\end{verbatim}

  A general function is used to update the database in both cases, based on the phrase and specific update function provided.

\begin{verbatim}

> updateBy :: ((SaxFCombination, FingeringTempo) -> 
>              TechniqueMap -> TechniqueMap) -> 
>             Music -> Integer -> Integer -> TechniqueMap -> TechniqueMap
> updateBy g m bpm bpMeas t = f (zip s ft) t 
>   where s :: [SaxFCombination]
>         s  = fingeringCombinations m
>         ft = fingeringTempos m bpm bpMeas
>         f :: [(SaxFCombination, FingeringTempo)] -> 
>              TechniqueMap -> TechniqueMap
>         f []     t' = t'
>         f (x:s') t' = f s' (g x t')

\end{verbatim}




