
\section{Typesetting}

While Haskore provides a very abstract and flexible representation of music, it is necessary to extend this representation somewhat to define a minimal representation of musical notation sufficient for the etude generation task described in this paper.  The {\tt Scores} module contains this Haskore extension (a Haskell data type), as well as a mechanism for typesetting scores encoded with that data type.

\begin{verbatim}

> module Scores where
> import Basics
> import Ratio
> import List
> import ComposersWorkbench
> import Miscellany

\end{verbatim}

\subsection{Score Representation}

To simplify the development of analysis and typesetting algorithms, a restricted musical vocabulary is used in {\tt Score}.  In this simple model, a {\tt Score} has only one time signature, tempo marking, and key throughout.  Also, a {\tt Score} must be monophonic, and not contain arbitrary tuples, although triplets are permitted.  {\tt Score} is essentially an extension of the Haskore data type {\tt Music}, with notation-specific features like keys and time signatures.

\begin{verbatim}

> data Score = Score {title :: Title, 
>                     timeSignature ::  TimeSignature, 
>                     beatsPerMinute :: Integer, 
>                     key :: Key, 
>                     music ::  Music}
>     deriving (Show, Eq)

> data Key =  Key PitchClass Mode
>     deriving (Show, Eq)

> data Mode =  Major | Minor
>     deriving (Show, Eq)

> type Title = String
> type TimeSignature = Ratio Int

> -- timeSigToBPMeas converts a time signature to 
> --   a number of beats per measure
> timeSigToBPMeas :: TimeSignature -> Integer
> timeSigToBPMeas = fromInt.numerator

> scoreWithTitle :: Title -> [Score] -> Score
> scoreWithTitle t s = (filter (\(Score t' _ _ _ _) -> t' == t) s) !! 0 

> -- phraseNum finds the nth phrase in a score
> phraseNum :: Integer -> Score -> Music
> phraseNum n (Score _ _ _ _ m) = (phrases m []) !! (fromInteger n)
>   where phrases :: Music -> [Music] -> [Music]
>         phrases m' acc = case m' of
>           (Phrase _ _)  -> acc ++ [m']
>           (m1 :+: m2)   -> phrases m2 (phrases m1 acc)
>           (_ :=: _)     -> error "polyphony not yet supported"
>           (Tempo _ m0)  -> phrases m0 acc
>           (Trans _ m0)  -> phrases m0 acc
>           (Instr _ m0)  -> phrases m0 acc
>           (Player _ m0) -> phrases m0 acc
>           _             -> acc 

\end{verbatim}

\subsection{LilyPond Scores}

The back-end of the typesetting subsystem is LilyPond, \cite{gltmt-m} an ``automated engraving system'' that is available on many platforms.  LilyPond has a command-line interface for compiling scores encoded in the LilyPond input format into several printable formats, including PostScript.  

The interface between the system described here and LilyPond is through the LilyPond input format.  This format is actually a rather substantial domain-specific language, implemented in Scheme, and only a subset of its functionality is required for this project.  This functionality is modelled internally with the {\tt LilyPondScore} data type.

\begin{verbatim}

> type LilyPondScore = [LilyPondSection] 

\end{verbatim}

In this (simplified) model, a {\tt LilyPondScore} is a list of sections, each of which may be a comment, header, version declaration, a ``relative octave'' mark, or a block of more primitive LilyPond expressions.  The header is primarily used to express the title of the piece, the version declaration specifies the version of LilyPond to which the input file conforms, and the ``relative octave'' mark tells the LilyPond interpreter to use the same octave for several consecutive notes. 

\begin{verbatim}

> data LilyPondSection = LPComment String
>                      | LPHeader String
>                      | LPVersion String
>                      | Relative 
>                      | LPBlock [LilyPondMark] 
>     deriving (Show, Eq)

\end{verbatim}

More importantly, the typographic elements for notes, barlines, ties, keys, and tempo markings are represented by the {\tt LilyPondMark} data type.  Each of these may appear anywhere within a given block of LilyPond expressions.

\begin{verbatim}

> data LilyPondMark = LPNote LPPitchClass LPDuration 
>                   | Tie
>                   | LPKey Key
>                   | MetronomeMark {beatLength :: LPDuration, 
>                                    beatsPerMinute' ::  Integer} 
>                   | FinalBar                   
>     deriving (Show, Eq)    

\end{verbatim}

  Pitches and are represented in the LilyPond input format by pitch class names modified with quote marks, the number of which denote the octave of the pitch.  Flats and sharps are represented by adding the suffixes ``es'' and ``is'', respectively, to the pitch class name.  Note durations are limited to the set of particular notehead glyphs that are available (for everything else, ties must be used).  1 denotes whole notes, 2 denotes half notes, 4 denotes quarter notes, etc.  One example of a note in the LilyPond input format is bes''16, which is a B-flat 16th note in the second octave of whichever clef is being used.

\begin{verbatim}

> type LPPitchClass = String 
> type LPDuration = String

\end{verbatim}

\subsection{Transforming Scores into LilyPond Scores}

  Typesetting a {\tt Score} involves several steps.   First, the {\tt Score} to be rendered is transformed into an internal representation of a sufficient subset of the LilyPond notation.  Then, this data structure is written to a file, using the LilyPond input format syntax.  Finally, the LilyPond shell command is invoked programmatically to produce a Portable Document Format (PDF) file, suitable to print on almost any operating system.

The {\tt score2LPScore} function converts {\tt Score}s to {\tt LilyPondScore}s.  Primarily, this is achieved by converting Haskore music values into sequences of {\tt LilyPondMarks}.  As mentionned previously some Haskore constructs like polyphony are omitted for simplicity in implementation, although it is possible to add this functionality later.  

\begin{verbatim}

> -- In the following, the quarternote beat is assumed 
> --   for now, but this eventually, should be derived from 
> --   the time signature
> score2LPScore :: Score -> LilyPondScore
> score2LPScore (Score t _ bpm k m) 
>   = [(LPComment "Generated from a Score by Scores.lhs."),
>      (LPHeader t),
>      LPBlock ((LPKey k):(MetronomeMark "4" bpm):b), 
>      (LPVersion "2.6.0")]
>   where b = ((music2NoteSeq m 0 4 1) ++ [FinalBar])

\end{verbatim}

  The conversion of {\tt Score}s to {\tt LilyPondScore}s is non-trivial, since Haskore note durations (in the interpretation used here) are measured in measures and may be of arbitrary length, but LilyPond note durations are limited to the durations of noteheads.  In LilyPond, all other durations must be composed of multiple notes with ties.  Haskore's Music data type, on the other hand, does not have a notion of ties at all.  

  While it is relatively straightforward to convert a duration like Haskore's {\tt 3\%4} into three quarter notes tied together, this is, of course, contrary to conventional music typesetting practice.  In this case, {\tt 3\%4} should be written as a dotted-half note, if it occurs on the first or second beat of the measure in 4/4 time.  Matters become even more complicated when a note does not begin or end on a beat.  For example, a {\tt 1\%8} rest followed by a {\tt 1\%4} note should be written as an eighth-note rest followed by two tied eighth-notes.

The process for converting durations is as follows.  First the duration is converted into a number of ticks, where one tick is one 32nd of a measure in length, and rounding is used as necessary.  

\begin{verbatim}

> durToTicks :: (Integral a) => Ratio a -> Integer
> durToTicks d 
>  =  if (isWholeNum ts)
>     then iRatioFloor ts
>     else error ("@todo cannot handle dur: " ++ (show d))
>   where ts :: Rational
>         ts = ((numTicksPerMeasure % 1) * (toRational d))

> numTicksPerMeasure :: Integer
> numTicksPerMeasure = 32

\end{verbatim}

Then, given the current position in the current measure, the duration (in ticks) is divided up into a set of durations, such that none of the new durations crosses a beat.  A recursive function, {\tt makeLPRests}, is used to create dotted notes and larger note values in some cases.  (Future versions of the system will include a revised version of this function which will work in all cases)

\begin{verbatim}
durToLPRests: converts a duration to a string of LP Rests
dur: duration, in ticks
pos: amount of the measure already used up, in ticks
tpb: ticksPerBeat

> durToLPRests :: Integer -> Integer -> Integer -> [LilyPondMark]
> durToLPRests 0   _   _   = []
> durToLPRests dur pos tpb 
>   = (makeLPRests newDur numTicksPerMeasure) ++ 
>     (durToLPRests (dur - newDur) (pos + newDur) tpb)
>  where nextBeatPos = ((pos `div` tpb) + 1) * tpb
>        endPos      = pos + dur
>        newDur      = if endPos > nextBeatPos
>                      then nextBeatPos - pos
>                      else dur

makeLPRests: creates notes with dotting where possible
d: a duration in ticks 
b: a base number of ticks 

> makeLPRests :: Integer -> Integer -> [LilyPondMark]
> makeLPRests 0 _ = []
> makeLPRests d b 
>  = if ((rem numTicksPerMeasure d) == 0)
>    then [lpRest (numTicksPerMeasure `div` d)]
>    else if d < b
>         then makeLPRests d halfB
>         else note : (makeLPRests newR b)
>  where r      = d - b
>        useDot = r >= halfB
>        note   = if useDot 
>                 then lpRest' ((show (numTicksPerMeasure `div` b)) 
>                               ++ ".")
>                 else lpRest (numTicksPerMeasure `div` b)
>        newR   = if useDot then (d - (b + halfB)) else r
>        halfB  = if odd b 
>                 then error ("Expected b to be even: " ++ (show b)) 
>                 else b `div` 2

\end{verbatim}

Finally, the sets of durations for each note or rest are converted into note or rest marks.  For notes, each note in the group is given a pitch class mark and it is tied to its neighbors in the group.

\begin{verbatim}

> -- measurePos is the amount of the measure already used up
> -- bpMeas: Beats Per measure
> -- ml: measure length
> music2NoteSeq :: Music -> Rational -> Integer -> Dur -> [LilyPondMark]
> music2NoteSeq m measurePos bpMeas ml 
>  = case m of
>    (Note p d _)  -> intersperse Tie (map (f p) (dRests d))
>    (Rest d)      -> dRests d 
>    (m1 :+: m2)   -> (music2NoteSeq m1 measurePos bpMeas ml) 
>                     ++ (music2NoteSeq m2 (measurePos + 
>                                           (durToRational (dur m1))) 
>                                       bpMeas ml)
>    (_ :=: _)     -> error "polyphony not yet supported"
>    (Tempo _ _)   -> error "tempo not yet supported"
>    (Trans i m')  -> music2NoteSeq 
>                       (transposeM i m') measurePos bpMeas ml
>    (Instr _ _)   -> error "Instr constructor not yet supported"
>    (Player _ _)  -> error "Player constructor not yet supported"
>    (Phrase _ m') -> music2NoteSeq m' measurePos bpMeas ml
>   where dRests d = durToLPRests 
>                      (durToTicks d) (durToTicks measurePos) tpb
>         tpb = if (rem numTicksPerMeasure bpMeas) == 0
>               then (numTicksPerMeasure `div` bpMeas)
>               else error ("num ticks per measure: " ++ 
>                           (show numTicksPerMeasure) ++ 
>                           " is incompatible with " ++
>                           "beats per measure: " ++ 
>                           (show bpMeas))
>         f p (LPNote _ d') = LPNote (lpPitchString p) d'

> lpRest d' = lpRest' (show d')

> lpRest' d = LPNote "r" d

> lpPitchString :: Pitch -> String
> lpPitchString (pc, o) = (toLPPitchClass pc) ++ octaveString
>   where octaveString = (replicate (o - 3) '\'') ++ (replicate (3 - o) ',')

> toLPPitchClass :: PitchClass -> LPPitchClass
> toLPPitchClass pc'
>   = case pc' of
>       Cf -> "ces";  C -> "c"; Cs -> "cis" 
>       Df -> "des";  D -> "d"; Ds -> "dis"
>       Ef -> "ees";  E -> "e"; Es -> "eis"
>       Ff -> "fes";  F -> "f"; Fs -> "fis"
>       Gf -> "ges";  G -> "g"; Gs -> "gis"
>       Af -> "aes";  A -> "a"; As -> "ais"
>       Bf -> "bes";  B -> "b"; Bs -> "bis" 

> modeString :: Mode -> String
> modeString m = case m of
>   Major -> "major"
>   Minor -> "minor"

\end{verbatim}

\subsection{Writing a LilyPond Score to a File}

Writing a {\tt LilyPondScore} is a simple matter of mapping {\tt LilyPondMark}s and {\tt LilyPondSection}s to their string representations, and concatenating all of those strings together.  The result is a valid LilyPond input script.
 
\begin{verbatim}

> renderLPScore :: LilyPondScore -> String
> renderLPScore sections = concat (map rs sections)
>  where rs s = (renderSection s) ++ "\n"
>        renderSection (LPComment c) = lpComment c
>        renderSection (LPHeader t)  = lpHeader t
>        renderSection (LPVersion v) = lpVersion v
>        renderSection Relative      = "\\relative"
>        renderSection (LPBlock marks) 
>          = "{" ++ 
>             (concat (map (\x -> (renderLPMark x) ++ " ") marks)) 
>             ++ "}"

> renderLPMark :: LilyPondMark -> String
> renderLPMark m = case m of
>   (LPKey (Key pc mode))   -> "\\key " ++ (toLPPitchClass pc) ++ " " 
>                              ++ "\\" ++ (modeString mode)
>   (MetronomeMark bl bpm) -> "\\tempo " ++ bl ++ "=" ++ (show bpm)
>   (LPNote p d)           -> p ++ d
>   Tie                    -> "~"
>   FinalBar               -> "\\bar \"|.\""

> lpComment text = "%{" ++ text ++ "%}"
> lpHeader title = lpSection "header" ("title = \"" ++ title ++ "\"")
> lpSection sectionName body = "\\" ++ sectionName ++ "{" ++ body ++ "}"
> lpVersion version = "\\version \"" ++ version ++ "\""

\end{verbatim}



